--[[
    LootMonitor 2.06
    Fix Bug.

    2.05
    Update toc.
    
    2.04
    Fix: Sometime Announce.Index isn't legal.
    
    2.03
    Add: Always Announcer the Looter Checkbox

    2.02
    Fix: CHAT_MSG_ADDON: LMPX, "end CorpseGUID NumQuality" instead "end CorpseGUID Num".
    Get Correct Quality Rank.

    2.01
    What's New?
    1. Rewrite Sync Function, use table replace single variables to store group's states, make code more readable.
    2. Make Addon work correct in party mode. Caused by Party Sequence, Only Announce Loots when Party Leader Installed this Addon.
    3. Compatible with older Version. (1.38, 1.39 and 1.40)
    4. Make GUI more pleasing to the eye. Now, Announcer in group will mark as Crimson, player who do not have this Addon will display as Grey.
    5. Add Single Disable Mode. You will not announce anything if all group member except you do not installed this Addon.
]]

local AddonName, Addon = ...

-- 主Frame（不可见）
Addon.Frame = CreateFrame("Frame")
-- 设置界面（界面逻辑放置于GUI.lua）
Addon.SetWindow = {}
Addon.GroupState = {} -- 40个状态栏
-- 本地化相关
local L = Addon.L

-- 相关变量
local Frame = Addon.Frame
Frame:Hide() -- 隐藏Frame，避免Update开销
-- 版本号（不再读取toc文件，/reload后版本号立刻生效）
Addon.Version = 2.08
local Version = Addon.Version
-- Addon MSG Prefix
local PrefixLM = "LMPX"
-- 设置相关
Addon.Config = {
    ["OutputToRaidWarning"] = false,  -- 输出至Raid_Warning频道（向下兼容）
    ["Quality"] = 3, -- 默认通报物品等级：蓝色（3）
    ["SingleDisable"] = false,
    ["ForceAnnounce"] = false,
}
local Config = Addon.Config
-- Lua方法加速
local t_insert = table.insert

-- WOW API加速
-- 普通函数
local UnitName = UnitName
local UnitGUID = UnitGUID
local UnitIsDead = UnitIsDead
local UnitInRaid = UnitInRaid
local IsInGroup = IsInGroup
local IsInRaid = IsInRaid
local UnitInBattleground = UnitInBattleground
local UnitIsGroupLeader = UnitIsGroupLeader
local UnitIsGroupAssistant = UnitIsGroupAssistant
local GetLootInfo = GetLootInfo
local GetLootSlotLink = GetLootSlotLink
local GetCVar = GetCVar
local SendChatMessage = SendChatMessage
-- 同步相关函数
local SendAddonMessage = C_ChatInfo.SendAddonMessage
local IsAddonMessagePrefixRegistered = C_ChatInfo.IsAddonMessagePrefixRegistered
local RegisterAddonMessagePrefix = C_ChatInfo.RegisterAddonMessagePrefix
-- 常量
local PlayerName = ""
local ItemQuality = {
    [1] = L[" Common+ Item(s)"],
    [2] = L[" Uncommon+ Item(s)"],
    [3] = L[" Rare+ Item(s)"],
    [4] = L[" Epic+ Item(s)"],
}
-- 控制变量
local Announce = {
    ["Rank"] = "NONE",
    ["Index"] = 0,
}
local LooterInfo = {}
local ItemIndex = 0
local GroupMemberStates = {}

-- 通用功能函数
-- 更新Table（合并）
function Addon:UpdateTable(Target, Source)
    for k, v in pairs(Source) do
        if type(v) == "table" then
            if type(Target[k]) == "table" then
                self:UpdateTable(Target[k], v)
            else
                Target[k] = self:UpdateTable({}, v)
            end
        elseif type(Target[k]) ~= "table" then
            Target[k] = v
        end
    end
    return Target
end

-- 操作函数
-- 升级
function Addon:DoUpdateCheck(VersionString)
    if tonumber(VersionString) > Version then
        print(L["<|cFFFF4500LM|r>Need Update your |cFFFF4500LootMonitor|r to newest one!"])
        print(L["|cFF4169E1Download Link|r https://www.curseforge.com/wow/addons/lootmonitor"])
    end
end
-- 检查
function Addon:DoCheck()
    if UnitInBattleground("player") then -- 战场禁用
        return
    end
    if IsInRaid() then
        SendAddonMessage(PrefixLM, "check "..tostring(Version), "raid")
    elseif IsInGroup() then
        SendAddonMessage(PrefixLM, "check "..tostring(Version), "party")
    end
    for i = 1, 40 do
        Addon.GroupState[i]:SetText("")
    end
end
-- 答复
function Addon:DoReply()
    if UnitInBattleground("player") then -- 战场禁用
        return
    end
    if IsInRaid() then
        SendAddonMessage(PrefixLM, "reply "..tostring(Version), "raid")
    elseif IsInGroup() then
        SendAddonMessage(PrefixLM, "reply "..tostring(Version), "party")
    end
end
-- 显示状态
function Addon:DoShowState()
    if next(GroupMemberStates) == nil then
        return
    end
    local AnnouncerColorPrefix = "|cFFDC143C"
    local NoAddonColorPrefix = "|cFF606060"
    local ColorSuffix = "|r"
    for i = 1, #GroupMemberStates do
        if GroupMemberStates[i] then
            local msg = GroupMemberStates[i].Name
            if i == Announce.Index then
                msg = AnnouncerColorPrefix..msg..ColorSuffix.." (v"..tostring(GroupMemberStates[i].Version)..")"
            elseif GroupMemberStates[i].Version == 0 then
                msg = NoAddonColorPrefix..msg.." ("..L["NoAddon"]..")"..ColorSuffix
            else
                msg = msg.." (v"..tostring(GroupMemberStates[i].Version)..")"
            end
            Addon.GroupState[i]:SetText(msg)
            -- Addon.GroupState[i]:Show()
        end
    end
end
-- 初始化团队信息
function Addon:InitGroupInfo()
    if UnitInBattleground("player") then -- 战场禁用
        return
    end
    GroupMemberStates = {}
    if IsInRaid() then -- 初始化Raid团队信息
        for i = 1, 40 do
            local u = "raid"..i
            local n = UnitName(u)
            if UnitName(u) then
                local state = {
                    ["Index"] = u,
                    ["Name"] = n,
                    ["Rank"] = "NONE",
                    ["LootMonitor"] = false,
                    ["Version"] = 0,
                }
                t_insert(GroupMemberStates, state)
            else
                break
            end
        end
    elseif IsInGroup() then -- 初始化队伍信息（Player放在最前）
        local PlayerRank = "NONE"
        if UnitIsGroupLeader("player") then
            PlayerRank = "LEADER"
            Announce.Rank = "LEADER"
            Announce.Index = 1
        end
        local state = {
            ["Index"] = "player",
            ["Name"] = PlayerName,
            ["Rank"] = PlayerRank,
            ["LootMonitor"] = true,
            ["Version"] = Version,
        }
        t_insert(GroupMemberStates, state)
        for i = 1, 4 do
            local u = "party"..i
            local n = UnitName(u)
            if UnitName(u) then
                local state = {
                    ["Index"] = u,
                    ["Name"] = n,
                    ["Rank"] = "NONE",
                    ["LootMonitor"] = false,
                    ["Version"] = 0,
                }
                t_insert(GroupMemberStates, state)
            else
                break
            end
        end
    end
end
-- 更新团队成员信息
function Addon:UpdateGroupInfo(SenderName, VersionString)
    if IsInRaid() then
        local SenderIndex = UnitInRaid(SenderName)
        if SenderIndex then
            local SenderRank = "NONE"
            if UnitIsGroupLeader(SenderName) then -- 信息发送者是团长
                SenderRank = "LEADER"
                Announce.Rank = "LEADER"
                Announce.Index = SenderIndex
            elseif UnitIsGroupAssistant(SenderName) then -- 信息发送者是助理
                SenderRank = "ASSISTANT"
                if Announce.Rank == "NONE" then
                    Announce.Rank = "ASSISTANT"
                    Announce.Index = SenderIndex
                elseif Announce.Rank == "ASSISTANT" then
                    if SenderIndex < Announce.Index then
                        Announce.Index = SenderIndex
                    end
                end
            else -- 普通团员
                if Announce.Rank == "NONE" then
                    if Announce.Index == 0 then
                        Announce.Index = SenderIndex
                    elseif SenderIndex < Announce.Index then
                        Announce.Index = SenderIndex
                    end
                end
            end
            local SenderVersion = tonumber(VersionString)
            GroupMemberStates[SenderIndex] = {
                ["Index"] = "raid"..SenderIndex,
                ["Name"] = SenderName,
                ["Rank"] = SenderRank,
                ["LootMonitor"] = true,
                ["Version"] = SenderVersion,
            }
        end
    elseif IsInGroup() then
        local SenderRank = "NONE"
        if UnitIsGroupLeader(SenderName) then
            SenderRank = "LEADER"
        end
        local SenderVersion = tonumber(VersionString)
        for i = 1, #GroupMemberStates do
            if GroupMemberStates[i].Name == SenderName then
                GroupMemberStates[i].Rank = SenderRank
                GroupMemberStates[i].LootMonitor = true
                GroupMemberStates[i].Version = SenderVersion
                if SenderRank == "LEADER" then
                    Announce.Rank = "LEADER"
                    Announce.Index = i
                end
                break
            end
        end
    end
end
-- 是否全团安装
function Addon:AllHaveLootMonitor()
    for i = 1, #GroupMemberStates do
        if not GroupMemberStates[i].LootMonitor then
            return false
        end
    end
    return true
end
-- 单人安装禁用
function Addon:SingleInstall()
    local count = 0
    for i = 1, #GroupMemberStates do
        if GroupMemberStates[i].LootMonitor then
            count = count + 1
            if count > 1 then
                return false
            end
        end
    end
    return true
end

-- 注册事件
Frame:RegisterEvent("ADDON_LOADED")
Frame:RegisterEvent("PLAYER_LOGOUT")
Frame:RegisterEvent("LOOT_OPENED")
Frame:RegisterEvent("CHAT_MSG_ADDON")
Frame:RegisterEvent("ZONE_CHANGED")
Frame:RegisterEvent("GROUP_ROSTER_UPDATE")

-- 事件分配
Frame:SetScript("OnEvent", function(self, event, ...)
    if type(self[event]) == "function" then
        return self[event](self, ...)
    end
end)

-- 事件处理
-- 加载处理
function Frame:ADDON_LOADED(Name)
    if Name ~= AddonName then
        return
    end
    -- 反注册ADDON_LOADED事件
    self:UnregisterEvent("ADDON_LOADED")
    -- 读取存档资料，如果没有LootMonitorDB.Config则新建
    if not LootMonitorDB.Config then
        LootMonitorDB = {
            ["Config"] = {},
        }
    end
    -- 用存档文件数据更新Config表
    Addon:UpdateTable(Config, LootMonitorDB.Config)
    -- 注册Prefix
    if not IsAddonMessagePrefixRegistered(PrefixLM) then
        RegisterAddonMessagePrefix(PrefixLM)
    end
    -- 反和谐相关
    if GetCVar("portal") == "CN" then
        ConsoleExec("portal KR")
        ConsoleExec("profanityFilter 0")
        ConsoleExec("overrideArchive 0")
    end
    -- 初始化常量
    PlayerName = UnitName("player")
    -- 初始化设定窗口
    Addon.SetWindow:Initialize()
    Addon:InitGroupInfo()
end
-- 退出处理（包括/reload）
function Frame:PLAYER_LOGOUT()
    -- 清空LootMonitorDB
    LootMonitorDB = {
        ["Config"] = {},
    }
    -- 用当前Config覆盖LootMonitorDB
    Addon:UpdateTable(LootMonitorDB.Config, Config)
end
-- 区域切换（清空Looter信息）
function Frame:ZONE_CHANGED()
    LooterInfo = {} -- 清空Looter信息
end
-- ADDON频道信息处理
function Frame:CHAT_MSG_ADDON(...)
    if UnitInBattleground("player") then -- 战场禁用
        return
    end
    if not IsInGroup() then -- 非组队情况下退出
        return
    end
    -- 获取事件信息
    local arg = {...}
    -- 团队和小队处理模式一致
    if arg[1] == PrefixLM then
        -- 切分信息，以空格为界
        local CMD, REST = arg[2]:match("(%S+)%s+(.-)$")
        -- 确定输出频道
        local OutputChannel = "party"
        if IsInRaid() then
            OutputChannel = "raid"
            if UnitIsGroupLeader("player") or UnitIsGroupAssistant("player") then
                if Config.OutputToRaidWarning then
                    OutputChannel = "raid_warning"
                end
            end
        end
        if CMD:lower() == "check" then -- 收到检查命令，返回reply信息
            -- 进行答复
            Addon:DoReply()
            -- 比较版本号，提示更新
            if REST then
                Addon:DoUpdateCheck(REST)
            end
        elseif CMD:lower() == "reply" or CMD:lower() == "start" then
            if arg[5] and REST then
                Addon:UpdateGroupInfo(arg[5], REST) -- 更新团队插件信息
                -- 比较版本号，提示更新
                if REST then
                    Addon:DoUpdateCheck(REST)
                end
            end
        elseif CMD:lower() == "target" then
            local TargetName, TargetGUID = REST:match("(%S+)%s(.-)$")
            if TargetGUID and Announce.Index ~= 0 then
                if not LooterInfo[TargetGUID] then
                    LooterInfo[TargetGUID] = arg[5]
                    ItemIndex = 1
                    if GroupMemberStates[Announce.Index].Name == UnitName("player") and Addon:AllHaveLootMonitor() then
                        SendChatMessage(string.format(L["<LM><%s> has Looted <%s>'s Corpse"], arg[5], TargetName), OutputChannel)
                    elseif GroupMemberStates[Announce.Index].Name == UnitName("player") and Config.ForceAnnounce then
                        SendChatMessage(L["<LM>Among all members who installed LootMonitor"], OutputChannel)
                        SendChatMessage(string.format(L["<LM><%s> has Looted <%s>'s Corpse"], arg[5], TargetName), OutputChannel)
                    else
                        SendChatMessage(string.format(L["<LM><%s>'s Corpse Opened"], TargetName), OutputChannel)
                    end
                end
            end
        elseif CMD:lower() == "loot" and ItemIndex ~= 0 then
            local TargetGUID, LootLink = REST:match("(%S+)%s+(.-)$")
            if LooterInfo[TargetGUID] == arg[5] and Announce.Index ~= 0 then
                if GroupMemberStates[Announce.Index].Name == PlayerName then
                    SendChatMessage(string.format("<LM>%d: %s", ItemIndex, LootLink), OutputChannel)
                end
                ItemIndex = ItemIndex + 1
            end
        elseif CMD:lower() == "end" and ItemIndex ~= 0 then
            local TargetGUID, LootsNum = REST:match("(%S+)%s+(.-)$")
            if LooterInfo[TargetGUID] == arg[5] and Announce.Index ~= 0 then
                if GroupMemberStates[Announce.Index].Name == PlayerName then
                    SendChatMessage(L["<LM>Total "]..LootsNum, OutputChannel)
                end
                ItemIndex = 0
            end
        end
    end
end
-- 队伍调整后重新初始化团队信息，并发送状态
function Frame:GROUP_ROSTER_UPDATE()
    if UnitInBattleground("player") then -- 战场禁用
        return
    end
    -- 初始化Announce表
    Announce.Rank = "NONE"
    Announce.Index = 0
    -- 初始化GroupMemberStates表
    Addon:InitGroupInfo()
    -- 同步版本信息
    if IsInRaid() then
        SendAddonMessage(PrefixLM, "reply "..tostring(Version), "raid")
    elseif IsInGroup() then
        SendAddonMessage(PrefixLM, "reply "..tostring(Version), "party")
    end
end
-- 发送Loot信息
function Frame:LOOT_OPENED(...)
    if UnitInBattleground("player") then -- 战场禁用
        return
    end
    if not IsInGroup() then -- 非组队状态禁用
        return
    end
    if Config.SingleDisable and Addon:SingleInstall() then -- 仅一人安装时禁用
        return
    end
    if not UnitIsDead("target") then -- 目标是否已死亡
        return
    end
    local loots = GetLootInfo() -- 获取Loot信息
    local msg = {} -- 初始化需要发送的Loot信息表
    -- 设置发送频道
    local channel = "party"
    if IsInRaid() then
        channel = "raid"
    end
    -- 根据设定的物品等级，填充物品信息表
    for i = 1, #loots do
        if loots[i] then
            if loots[i].quantity ~= 0 then
                if loots[i].quality >= Config.Quality then
                    t_insert(msg, GetLootSlotLink(i))
                end
            end
        end
    end
    -- 发送物品信息
    if #msg > 0 then
        SendAddonMessage(PrefixLM, "target " .. UnitName("target") .. " " .. UnitGUID("target"), channel)
        for i = 1, #msg do
            SendAddonMessage(PrefixLM, "loot " .. UnitGUID("target") .. " " .. msg[i], channel)
        end
        SendAddonMessage(PrefixLM, "end " .. UnitGUID("target") .. " " .. tostring(#msg)..ItemQuality[Config.Quality], channel)
    end
end

-- 注册slash命令
SLASH_LMC1 = "/lootmonitor"
SLASH_LMC2 = "/lm"
-- 处理slash命令
SlashCmdList["LMC"] = function()
    if Addon.SetWindow:IsShown() then
        Addon.SetWindow:Hide()
    else
        Addon.SetWindow:Show()
    end
end